#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='cpmd_general_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='cpmd.general.nomadmetainfo.json'))


class x_cpmd_section_start_information(MSection):
    '''
    Contains information about the starting conditions for this run
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_start_information'))

    x_cpmd_start_datetime = Quantity(
        type=str,
        shape=[],
        description='''
        CPMD run start time and date
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_start_datetime'))

    x_cpmd_input_filename = Quantity(
        type=str,
        shape=[],
        description='''
        CPMD input file name.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_input_filename'))

    x_cpmd_compilation_date = Quantity(
        type=str,
        shape=[],
        description='''
        CPMD compilation date.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_compilation_date'))

    x_cpmd_process_id = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The process id for this calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_process_id'))

    x_cpmd_run_user_name = Quantity(
        type=str,
        shape=[],
        description='''
        The user who launched this calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_run_user_name'))

    x_cpmd_run_host_name = Quantity(
        type=str,
        shape=[],
        description='''
        The host on which this calculation was made on.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_run_host_name'))


class x_cpmd_section_run_type_information(MSection):
    '''
    Contains information about the run type.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_run_type_information'))

    x_cpmd_time_step_ions = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The time step for ions.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_time_step_ions'))

    x_cpmd_time_step_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The time step for electrons.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_time_step_electrons'))

    x_cpmd_geo_opt_method = Quantity(
        type=str,
        shape=[],
        description='''
        The geometry optimization method.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_method'))

    x_cpmd_max_steps = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The maximum number of steps requested. In MD, this is the number of MD steps, in
        single point calculations this is the number of scf cycles, in geometry
        optimization this is the number of optimization steps.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_max_steps'))

    x_cpmd_ion_temperature_control = Quantity(
        type=str,
        shape=[],
        description='''
        The temperature control method for ion dynamics.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_ion_temperature_control'))


class x_cpmd_section_xc_information(MSection):
    '''
    Contains information about the exchange-correlation functional.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_xc_information'))


class x_cpmd_section_system_information(MSection):
    '''
    Contains information about the system.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_system_information'))


class x_cpmd_section_pseudopotential_information(MSection):
    '''
    Contains information about the pseudopotentials.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_pseudopotential_information'))


class x_cpmd_section_atom_kinds(MSection):
    '''
    Contains information about the atomic kinds present in the calculation.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_atom_kinds'))

    x_cpmd_section_atom_kind = SubSection(
        sub_section=SectionProxy('x_cpmd_section_atom_kind'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_atom_kind'))


class x_cpmd_section_atom_kind(MSection):
    '''
    Contains information about one atomic kind.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_atom_kind'))

    x_cpmd_atom_kind_label = Quantity(
        type=str,
        shape=[],
        description='''
        The label of the atomic kind.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_atom_kind_label'))

    x_cpmd_atom_kind_mass = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mass of the atomic kind.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_atom_kind_mass'))

    x_cpmd_atom_kind_raggio = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The width of the ionic charge distribution (RAGGIO) of the atomic kind.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_atom_kind_raggio'))

    x_cpmd_atom_kind_nlcc = Quantity(
        type=str,
        shape=[],
        description='''
        The nonlinear core correction (NLCC) of the atomic kind.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_atom_kind_nlcc'))

    x_cpmd_atom_kind_pseudopotential_l = Quantity(
        type=str,
        shape=[],
        description='''
        The angular part of the pseudopotential for the atomic kind.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_atom_kind_pseudopotential_l'))

    x_cpmd_atom_kind_pseudopotential_type = Quantity(
        type=str,
        shape=[],
        description='''
        The type of the pseudopotential for the atomic kind.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_atom_kind_pseudopotential_type'))


class x_cpmd_section_supercell(MSection):
    '''
    Contains information about the supercell.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_supercell'))

    x_cpmd_cell_symmetry = Quantity(
        type=str,
        shape=[],
        description='''
        The symmetry of the cell.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_cell_symmetry'))

    x_cpmd_cell_lattice_constant = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The cell lattice constant.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_cell_lattice_constant'))

    x_cpmd_cell_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The cell volume.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_cell_volume'))

    x_cpmd_cell_dimension = Quantity(
        type=str,
        shape=[],
        description='''
        The cell dimension.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_cell_dimension'))

    x_cpmd_lattice_vector_A1 = Quantity(
        type=str,
        shape=[],
        description='''
        Lattice vector A1
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_lattice_vector_A1'))

    x_cpmd_lattice_vector_A2 = Quantity(
        type=str,
        shape=[],
        description='''
        Lattice vector A2
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_lattice_vector_A2'))

    x_cpmd_lattice_vector_A3 = Quantity(
        type=str,
        shape=[],
        description='''
        Lattice vector A3
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_lattice_vector_A3'))

    x_cpmd_reciprocal_lattice_vector_B1 = Quantity(
        type=str,
        shape=[],
        description='''
        Reciprocal lattice vector B1
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_reciprocal_lattice_vector_B1'))

    x_cpmd_reciprocal_lattice_vector_B2 = Quantity(
        type=str,
        shape=[],
        description='''
        Reciprocal lattice vector B2
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_reciprocal_lattice_vector_B2'))

    x_cpmd_reciprocal_lattice_vector_B3 = Quantity(
        type=str,
        shape=[],
        description='''
        Reciprocal lattice vector B3
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_reciprocal_lattice_vector_B3'))

    x_cpmd_real_space_mesh = Quantity(
        type=str,
        shape=[],
        description='''
        Number of points in the real space mesh.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_real_space_mesh'))

    x_cpmd_wave_function_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Place wave cutoff energy for wave function.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_wave_function_cutoff'))

    x_cpmd_density_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Place wave cutoff energy for density.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_density_cutoff'))

    x_cpmd_number_of_planewaves_density = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of plane waves for density cutoff.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_number_of_planewaves_density'))

    x_cpmd_number_of_planewaves_wave_function = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of plane waves for wave_function cutoff.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_number_of_planewaves_wave_function'))


class x_cpmd_section_wave_function_initialization(MSection):
    '''
    Contains information about the wave function initialization
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_wave_function_initialization'))


class x_cpmd_section_scf(MSection):
    '''
    Contains information about self-consistent field calculation
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_scf'))

    x_cpmd_section_scf_iteration = SubSection(
        sub_section=SectionProxy('x_cpmd_section_scf_iteration'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_scf_iteration'))


class x_cpmd_section_scf_iteration(MSection):
    '''
    Contains information about the self-consistent field iteration within a wavefunction
    optimization.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_scf_iteration'))

    x_cpmd_scf_nfi = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The scf step number (NFI).
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_scf_nfi'))

    x_cpmd_scf_gemax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Largest off-diagonal component (GEMAX) during SCF step.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_scf_gemax'))

    x_cpmd_scf_cnorm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average of the off-diagonal components (CNORM) during SCF step.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_scf_cnorm'))

    x_cpmd_scf_etot = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The total energy (ETOT) during SCF step.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_scf_etot'))

    x_cpmd_scf_detot = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The difference in total energy to the previous SCF energy (DETOT).
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_scf_detot'))

    x_cpmd_scf_tcpu = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The CPU time used during SCF step (TCPU).
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_scf_tcpu'))


class x_cpmd_section_final_results(MSection):
    '''
    The final results after a single point calculation.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_final_results'))


class x_cpmd_section_geo_opt_initialization(MSection):
    '''
    Geometry optimization initialization information.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_geo_opt_initialization'))

    x_cpmd_total_number_of_molecular_structures = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Total number of molecular structures.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_total_number_of_molecular_structures'))

    x_cpmd_initialized_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        The initialized positions for geometry optimization. The ith row corresponds to
        the position for atom number i.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_initialized_positions'))

    x_cpmd_initialized_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        The initialized forces for geometry optimization. The ith row corresponds to the
        force for atom number i.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_initialized_forces'))

    x_cpmd_geo_opt_initialization_time = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Time for initialization.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_initialization_time'))


class x_cpmd_section_geo_opt_step(MSection):
    '''
    Contains information for a single geometry optimization step.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_geo_opt_step'))

    x_cpmd_geo_opt_step_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        The positions from a geometry optimization step. The ith row corresponds to the
        position for atom number i.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_step_positions'))

    x_cpmd_geo_opt_step_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        description='''
        The forces from a geometry optimization step. The ith row corresponds to the force
        for atom number i.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_step_forces'))

    x_cpmd_geo_opt_step_total_number_of_scf_steps = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Total number of SCF steps at the end of this geometry optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_step_total_number_of_scf_steps'))

    x_cpmd_geo_opt_step_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Geometry optimization step number.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_step_number'))

    x_cpmd_geo_opt_step_gnmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The largest absolute component of the force on any atom (GNMAX).
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_step_gnmax'))

    x_cpmd_geo_opt_step_gnorm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average force on the atoms (GNORM).
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_step_gnorm'))

    x_cpmd_geo_opt_step_cnstr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The largest absolute component of a constraint force on the atoms (CNSTR).
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_step_cnstr'))

    x_cpmd_geo_opt_step_etot = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The total energy at the end of a geometry optimization step (ETOT).
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_step_etot'))

    x_cpmd_geo_opt_step_detot = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The difference in total energy to the previous geometry optimization step (DETOT).
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_step_detot'))

    x_cpmd_geo_opt_step_tcpu = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The CPU time used during geometry optimization step (TCPU).
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_step_tcpu'))

    x_cpmd_section_geo_opt_scf_iteration = SubSection(
        sub_section=SectionProxy('x_cpmd_section_geo_opt_scf_iteration'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_geo_opt_scf_iteration'))


class x_cpmd_section_geo_opt_scf_iteration(MSection):
    '''
    Contains information about the self-consistent field iteration within a geometry
    optimization step.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_geo_opt_scf_iteration'))

    x_cpmd_geo_opt_scf_nfi = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The scf step number (NFI) within geometry optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_scf_nfi'))

    x_cpmd_geo_opt_scf_gemax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Largest off-diagonal component (GEMAX) during SCF step within geometry
        optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_scf_gemax'))

    x_cpmd_geo_opt_scf_cnorm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Average of the off-diagonal components (CNORM) during SCF step within geometry
        optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_scf_cnorm'))

    x_cpmd_geo_opt_scf_etot = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The total energy (ETOT) during SCF step within geometry optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_scf_etot'))

    x_cpmd_geo_opt_scf_detot = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The difference in total energy to the previous SCF energy (DETOT) within geometry
        optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_scf_detot'))

    x_cpmd_geo_opt_scf_tcpu = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The CPU time used during SCF step (TCPU) within geometry optimization step.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_geo_opt_scf_tcpu'))


class x_cpmd_section_md_initialization(MSection):
    '''
    Molecular dynamics initialization information.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_md_initialization'))


class x_cpmd_section_md_averaged_quantities(MSection):
    '''
    Averaged quantities from a MD calculation.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_md_averaged_quantities'))

    x_cpmd_electron_kinetic_energy_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean electron kinetic energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_electron_kinetic_energy_mean'))

    x_cpmd_electron_kinetic_energy_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The standard deviation of electron kinetic energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_electron_kinetic_energy_std'))

    x_cpmd_ionic_temperature_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean ionic temperature.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_ionic_temperature_mean'))

    x_cpmd_ionic_temperature_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The standard deviation of ionic temperature.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_ionic_temperature_std'))

    x_cpmd_density_functional_energy_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean density functional energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_density_functional_energy_mean'))

    x_cpmd_density_functional_energy_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The standard deviation of density functional energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_density_functional_energy_std'))

    x_cpmd_classical_energy_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean classical energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_classical_energy_mean'))

    x_cpmd_classical_energy_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The standard deviation of classical energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_classical_energy_std'))

    x_cpmd_conserved_energy_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean conserved energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_conserved_energy_mean'))

    x_cpmd_conserved_energy_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The standard deviation of conserved energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_conserved_energy_std'))

    x_cpmd_nose_energy_electrons_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean Nosé energy for electrons.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_nose_energy_electrons_mean'))

    x_cpmd_nose_energy_electrons_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The standard deviation of Nosé energy for elctrons.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_nose_energy_electrons_std'))

    x_cpmd_nose_energy_ions_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean Nosé energy for ions.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_nose_energy_ions_mean'))

    x_cpmd_nose_energy_ions_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The standard deviation of Nosé energy for ions.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_nose_energy_ions_std'))

    x_cpmd_constraints_energy_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean constrains energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_constraints_energy_mean'))

    x_cpmd_constraints_energy_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The standard deviation of constraints energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_constraints_energy_std'))

    x_cpmd_restraints_energy_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean restraints energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_restraints_energy_mean'))

    x_cpmd_restraints_energy_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The standard deviation of restraints energy.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_restraints_energy_std'))

    x_cpmd_ion_displacement_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean ion displacement.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_ion_displacement_mean'))

    x_cpmd_ion_displacement_std = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The standard deviation of ion displacement.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_ion_displacement_std'))

    x_cpmd_cpu_time_mean = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The mean cpu time.
        ''',
        a_legacy=LegacyDefinition(name='x_cpmd_cpu_time_mean'))


class x_cpmd_section_timing(MSection):
    '''
    Contains information about the timings.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_timing'))


class x_cpmd_section_end_information(MSection):
    '''
    Contains information printed at the end of a calculation.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_cpmd_section_end_information'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_cpmd_section_start_information = SubSection(
        sub_section=SectionProxy('x_cpmd_section_start_information'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_start_information'))

    x_cpmd_section_run_type_information = SubSection(
        sub_section=SectionProxy('x_cpmd_section_run_type_information'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_run_type_information'))

    x_cpmd_section_system_information = SubSection(
        sub_section=SectionProxy('x_cpmd_section_system_information'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_system_information'))

    x_cpmd_section_supercell = SubSection(
        sub_section=SectionProxy('x_cpmd_section_supercell'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_supercell'))

    x_cpmd_section_wave_function_initialization = SubSection(
        sub_section=SectionProxy('x_cpmd_section_wave_function_initialization'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_wave_function_initialization'))

    x_cpmd_section_md_initialization = SubSection(
        sub_section=SectionProxy('x_cpmd_section_md_initialization'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_md_initialization'))

    x_cpmd_section_md_averaged_quantities = SubSection(
        sub_section=SectionProxy('x_cpmd_section_md_averaged_quantities'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_md_averaged_quantities'))

    x_cpmd_section_timing = SubSection(
        sub_section=SectionProxy('x_cpmd_section_timing'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_timing'))

    x_cpmd_section_end_information = SubSection(
        sub_section=SectionProxy('x_cpmd_section_end_information'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_end_information'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_cpmd_section_xc_information = SubSection(
        sub_section=SectionProxy('x_cpmd_section_xc_information'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_xc_information'))

    x_cpmd_section_pseudopotential_information = SubSection(
        sub_section=SectionProxy('x_cpmd_section_pseudopotential_information'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_pseudopotential_information'))

    x_cpmd_section_atom_kinds = SubSection(
        sub_section=SectionProxy('x_cpmd_section_atom_kinds'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_atom_kinds'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_cpmd_section_scf = SubSection(
        sub_section=SectionProxy('x_cpmd_section_scf'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_scf'))

    x_cpmd_section_final_results = SubSection(
        sub_section=SectionProxy('x_cpmd_section_final_results'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_final_results'))


class section_frame_sequence(public.section_frame_sequence):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_frame_sequence'))

    x_cpmd_section_geo_opt_initialization = SubSection(
        sub_section=SectionProxy('x_cpmd_section_geo_opt_initialization'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_geo_opt_initialization'))

    x_cpmd_section_geo_opt_step = SubSection(
        sub_section=SectionProxy('x_cpmd_section_geo_opt_step'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_cpmd_section_geo_opt_step'))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
The archive API of the nomad@FAIRDI APIs. This API is about serving processed
(parsed and normalized) calculation data in nomad's *meta-info* format.
'''

from typing import Dict, Any
from io import BytesIO
from flask import request, g
from flask_restplus import abort, Resource, fields
import orjson
import urllib.parse

from nomad.files import UploadFiles, Restricted
from nomad.archive import (
    query_archive, ArchiveQueryError, filter_archive, read_partial_archives_from_mongo,
    compute_required_with_referenced)
from nomad import search, config

from .. import common
from .auth import authenticate, create_authorization_predicate
from .api import api
from .common import calc_route, streamed_zipfile, search_model, add_search_parameters,\
    apply_search_parameters


ns = api.namespace(
    'archive',
    description='Access archive data and archive processing logs.')


@calc_route(ns, '/logs')
class ArchiveCalcLogResource(Resource):
    @api.doc('get_archive_logs')
    @api.response(404, 'The upload or calculation does not exist')
    @api.response(401, 'Not authorized to access the data.')
    @api.response(200, 'Archive data send')
    @authenticate(signature_token=True)
    def get(self, upload_id, calc_id):
        '''
        Get calculation processing log.

        Calcs are references via *upload_id*, *calc_id* pairs.
        '''
        archive_id = '%s/%s' % (upload_id, calc_id)

        upload_files = UploadFiles.get(
            upload_id, is_authorized=create_authorization_predicate(upload_id, calc_id))

        if upload_files is None:
            abort(404, message='Upload %s does not exist.' % upload_id)

        try:
            with upload_files.read_archive(calc_id) as archive:
                return [entry.to_dict() for entry in archive[calc_id]['processing_logs']]

        except Restricted:
            abort(401, message='Not authorized to access %s/%s.' % (upload_id, calc_id))
        except KeyError:
            abort(404, message='Calculation %s does not exist.' % archive_id)


@calc_route(ns)
class ArchiveCalcResource(Resource):
    @api.doc('get_archive_calc')
    @api.response(404, 'The upload or calculation does not exist')
    @api.response(401, 'Not authorized to access the data.')
    @api.response(200, 'Archive data send', headers={'Content-Type': 'application/json'})
    @authenticate(signature_token=True)
    def get(self, upload_id, calc_id):
        '''
        Get calculation data in archive form.

        Calcs are references via *upload_id*, *calc_id* pairs.
        '''
        archive_id = '%s/%s' % (upload_id, calc_id)

        upload_files = UploadFiles.get(
            upload_id, is_authorized=create_authorization_predicate(upload_id, calc_id))

        if upload_files is None:
            abort(404, message='Archive %s does not exist.' % upload_id)

        try:
            with upload_files.read_archive(calc_id) as archive:
                return {
                    key: value
                    for key, value in archive[calc_id].to_dict().items()
                    if key != 'processing_logs'}

        except Restricted:
            abort(401, message='Not authorized to access %s/%s.' % (upload_id, calc_id))
        except KeyError:
            abort(404, message='Calculation %s does not exist.' % archive_id)


_archive_download_parser = api.parser()
add_search_parameters(_archive_download_parser)
_archive_download_parser.add_argument(
    name='compress', type=bool, help='Use compression on .zip files, default is not.',
    location='args')


@ns.route('/download')
class ArchiveDownloadResource(Resource):
    manifest_quantities = ['upload_id', 'calc_id', 'external_id', 'raw_id', 'pid', 'calc_hash']

    @api.doc('archive_download')
    @api.response(400, 'Invalid requests, e.g. wrong owner type or bad search parameters')
    @api.expect(_archive_download_parser, validate=True)
    @api.response(200, 'File(s) send', headers={'Content-Type': 'application/zip'})
    @authenticate(signature_token=True)
    def get(self):
        '''
        Get calculation data in archive form from all query results.

        See ``/repo`` endpoint for documentation on the search
        parameters.

        Zip files are streamed; instead of 401 errors, the zip file will just not contain
        any files that the user is not authorized to access.

        The zip file will contain a ``manifest.json`` with the repository meta data.
        '''
        try:
            args = _archive_download_parser.parse_args()
            compress = args.get('compress', False)
        except Exception:
            abort(400, message='bad parameter types')

        search_request = search.SearchRequest()
        apply_search_parameters(search_request, args)
        search_request.include('calc_id', 'upload_id', 'mainfile')

        calcs = search_request.execute_scan(
            order_by='upload_id',
            size=config.services.download_scan_size,
            scroll=config.services.download_scan_timeout)

        def generator():
            try:
                manifest = {}
                upload_files = None

                for entry in calcs:
                    upload_id = entry['upload_id']
                    calc_id = entry['calc_id']

                    manifest = {
                        calc_id: {
                            key: entry[key]
                            for key in ArchiveDownloadResource.manifest_quantities
                            if entry.get(key) is not None
                        }
                    }

                    if upload_files is None or upload_files.upload_id != upload_id:
                        if upload_files is not None:
                            upload_files.close()

                        upload_files = UploadFiles.get(
                            upload_id, create_authorization_predicate(upload_id))

                        if upload_files is None:
                            common.logger.error('upload files do not exist', upload_id=upload_id)
                            continue

                    upload_files._is_authorized = create_authorization_predicate(
                        upload_id=upload_id, calc_id=calc_id)
                    with upload_files.read_archive(calc_id) as archive:
                        f = BytesIO(orjson.dumps(
                            archive[calc_id].to_dict(),
                            option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS))

                        yield (
                            '%s.%s' % (calc_id, 'json'), calc_id, manifest,
                            lambda calc_id: f,
                            lambda calc_id: f.getbuffer().nbytes)

                if upload_files is not None:
                    upload_files.close()

            except Exception as e:
                common.logger.error(
                    'unexpected error while streaming raw data from query',
                    exc_info=e,
                    query=urllib.parse.urlencode(request.args, doseq=True))

        return streamed_zipfile(
            generator(), zipfile_name='nomad_archive.zip', compress=compress, manifest=dict())


_archive_query_model = api.inherit('ArchiveSearch', search_model, {
    'required': fields.Raw(description='A dictionary that defines what archive data to retrive.'),
    'query_schema': fields.Raw(description='Deprecated, use required instead.'),
    'raise_errors': fields.Boolean(
        description='Return 404 on missing archives or 500 on other errors instead of skipping the entry.')
})


@ns.route('/query')
class ArchiveQueryResource(Resource):
    @api.doc('post_archive_query')
    @api.response(400, 'Invalid requests, e.g. wrong owner type or bad search parameters')
    @api.response(401, 'Not authorized to access the data.')
    @api.response(404, 'The upload or calculation does not exist')
    @api.expect(_archive_query_model)
    @api.marshal_with(_archive_query_model, skip_none=True, code=200, description='Archive search results sent')
    @authenticate()
    def post(self):
        '''
        Post a query schema and return it filled with archive data.

        See ``/repo`` endpoint for documentation on the search
        parameters.

        This endpoint uses pagination (see /repo) or id aggregation to handle large result
        sets over multiple requests.
        Use aggregation.after and aggregation.per_page to request a
        certain page with id aggregation.

        The actual data are in results and a supplementary python code (curl) to
        execute search is in python (curl).
        '''
        try:
            data_in = request.get_json()
            aggregation = data_in.get('aggregation', None)

            pagination = data_in.get('pagination', {})
            page = pagination.get('page', 1)
            per_page = pagination.get('per_page', 10)

            query = data_in.get('query', {})

            query_expression = {key: val for key, val in query.items() if '$' in key}

            required: Dict[str, Any] = None
            if 'required' in data_in:
                required = data_in.get('required')
            else:
                required = data_in.get('query_schema', '*')

            raise_errors = data_in.get('raise_errors', False)

        except Exception:
            abort(400, message='bad parameter types')

        if not (page >= 1 and per_page > 0):
            abort(400, message='invalid pagination')

        search_request = search.SearchRequest()
        if g.user is not None:
            search_request.owner('all', user_id=g.user.user_id)
        else:
            search_request.owner('all')

        apply_search_parameters(search_request, query)
        if not aggregation:
            search_request.include(
                'calc_id', 'upload_id', 'with_embargo', 'published', 'parser_name')

        if query_expression:
            try:
                search_request.query_expression(query_expression)
            except AssertionError as e:
                abort(400, str(e))

        try:
            if aggregation:
                results = search_request.execute_aggregated(
                    after=aggregation.get('after'), per_page=aggregation.get('per_page', 1000),
                    includes=['with_embargo', 'published', 'parser_name'])

            else:
                results = search_request.execute_paginated(
                    per_page=per_page, page=page, order_by='upload_id')

        except KeyError as e:
            abort(400, 'The query part contained an unknown quantity or section: %s' % str(e))

        try:
            required_with_references = compute_required_with_referenced(required)
        except KeyError as e:
            abort(
                400,
                message='The required part contained an unknown quantity or section: %s' % str(e))

        data = []
        calcs = results['results']
        upload_files = None
        current_upload_id = None
        archive_is_complete = False
        if required_with_references is not None:
            archive_is_complete = True
            required = required_with_references

        for entry in calcs:
            with_embargo = entry['with_embargo']

            upload_id = entry['upload_id']
            calc_id = entry['calc_id']

            if upload_files is None or current_upload_id != upload_id:
                if upload_files is not None:
                    upload_files.close()

                upload_files = UploadFiles.get(
                    upload_id, create_authorization_predicate(upload_id))

                if upload_files is None:
                    return []

                if archive_is_complete:
                    upload_calc_ids = [
                        calc['calc_id'] for calc in calcs if calc['upload_id'] == upload_id]
                    upload_partial_archives = read_partial_archives_from_mongo(
                        upload_calc_ids, as_dict=True)

                current_upload_id = upload_id

            # TODO we are either just use the partial from mongo or read the whole required
            # from the mgs-pack archive on top.
            # Ideally, we would only read whats left from the msg-pack archive and merge.
            if archive_is_complete:
                try:
                    partial_archive = upload_partial_archives[calc_id]
                    partial_archive = filter_archive(
                        required, partial_archive, transform=lambda e: e)

                    data.append({
                        'calc_id': calc_id,
                        'parser_name': entry['parser_name'],
                        'archive': partial_archive})
                    continue
                except KeyError:
                    pass
                except ArchiveQueryError as e:
                    abort(400, str(e))
                except Exception as e:
                    common.logger.error(
                        str(e), upload_id=upload_id, calc_id=calc_id, exc_info=e)

            if with_embargo:
                access = 'restricted'
                upload_files._is_authorized = create_authorization_predicate(
                    upload_id=upload_id, calc_id=calc_id)
            else:
                access = 'public'

            try:
                with upload_files.read_archive(calc_id, access) as archive:
                    data.append({
                        'calc_id': calc_id,
                        'parser_name': entry['parser_name'],
                        'archive': query_archive(archive, {calc_id: required})[calc_id]
                    })
            except ArchiveQueryError as e:
                abort(400, str(e))
            except KeyError:
                if raise_errors:
                    abort(404, 'Archive for entry %s does not exist' % calc_id)
                # We simply skip this entry
                pass
            except Restricted:
                # this should not happen
                common.logger.error(
                    'supposedly unreachable code', upload_id=upload_id, calc_id=calc_id)
            except Exception as e:
                if raise_errors:
                    raise e
                common.logger.error(
                    str(e), upload_id=upload_id, calc_id=calc_id, exc_info=e)

        if upload_files is not None:
            upload_files.close()

        # assign archive data to results
        results['results'] = data

        return results, 200



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np

from nomad.normalizing.normalizer import Normalizer
from nomad.datamodel import EntryArchive
from nomad.datamodel.metainfo.public import Workflow, GeometryOptimization, Phonon, Elastic,\
    MolecularDynamics, SinglePoint


def resolve_difference(values):
    delta_values = None

    values = [v for v in values if v is not None]
    for n in range(-1, -len(values), -1):
        delta_values = abs(values[n] - values[n - 1])
        if delta_values != 0.0:
            break

    return delta_values


class SinglePointNormalizer(Normalizer):
    def __init__(self, entry_archive):
        super().__init__(entry_archive)

    def normalize(self):
        self.section = self.entry_archive.section_workflow.section_single_point
        if not self.section:
            self.section = self.entry_archive.section_workflow.m_create(SinglePoint)

        if not self.section.single_point_calculation_method:
            try:
                method = self.section_run.section_method[-1]
                self.section.single_point_calculation_method = method.electronic_structure_method
            except Exception:
                pass

        scc = self.section_run.section_single_configuration_calculation
        if not scc:
            return

        if not self.section.number_of_scf_steps:
            self.section.number_of_scf_steps = len(scc[-1].section_scf_iteration)

        energies = [scf.energy_total_scf_iteration for scf in scc[-1].section_scf_iteration]
        delta_energy = resolve_difference(energies)
        if not self.section.final_scf_energy_difference and delta_energy is not None:
            self.section.final_scf_energy_difference = delta_energy

        if not self.section.is_converged and delta_energy is not None:
            try:
                threshold = self.section_run.section_method[-1].scf_threshold_energy_change
                self.section.is_converged = bool(delta_energy <= threshold)
            except Exception:
                pass

        if not self.section.with_density_of_states:
            self.section.with_density_of_states = len(scc[-1].section_dos) > 0

        if not self.section.with_bandstructure:
            self.section.with_bandstructure = len(scc[-1].section_k_band) > 0

        if not self.section.with_eigenvalues:
            self.section.with_eigenvalues = len(scc[-1].section_eigenvalues) > 0

        if not self.section.with_volumetric_data:
            self.section.with_volumetric_data = len(scc[-1].section_volumetric_data) > 0

        if not self.section.with_excited_states:
            self.section.with_excited_states = len(scc[-1].section_excited_states) > 0


class GeometryOptimizationNormalizer(Normalizer):
    def __init__(self, entry_archive):
        super().__init__(entry_archive)

    def _to_numpy_array(self, quantity):
        return np.array(quantity.m if quantity is not None else quantity)

    def _get_geometry_optimization_type(self):
        sec_system = self.section_run.section_system
        if not sec_system:
            return

        def compare_cell(cell1, cell2):
            if (cell1 == cell2).all():
                return None
            else:
                cell1_normed = cell1 / np.linalg.norm(cell1)
                cell2_normed = cell2 / np.linalg.norm(cell2)
                if (cell1_normed == cell2_normed).all():
                    return 'cell_volume'
                else:
                    return 'cell_shape'

        if len(sec_system) < 2:
            return 'static'

        else:
            cell_init = sec_system[0].lattice_vectors
            cell_final = sec_system[-1].lattice_vectors
            if cell_init is None:
                cell_init = sec_system[0].simulation_cell
            if cell_final is None:
                cell_final = sec_system[-1].simulation_cell

            cell_init = self._to_numpy_array(cell_init)
            cell_final = self._to_numpy_array(cell_final)

            cell_relaxation = compare_cell(cell_init, cell_final)

            if cell_relaxation is not None:
                return cell_relaxation

            atom_pos_init = self._to_numpy_array(sec_system[0].atom_positions)
            atom_pos_final = self._to_numpy_array(sec_system[-1].atom_positions)

            if (atom_pos_init == atom_pos_final).all():
                return 'static'

            return 'ionic'

    def normalize(self):
        self.section = self.entry_archive.section_workflow.section_geometry_optimization
        if not self.section:
            self.section = self.entry_archive.section_workflow.m_create(GeometryOptimization)

        if not self.section.geometry_optimization_type:
            try:
                geometry_optimization_type = self._get_geometry_optimization_type()
                self.section.geometry_optimization_type = geometry_optimization_type
            except Exception:
                pass

        if not self.section.optimization_steps:
            scc = self.section_run.section_single_configuration_calculation
            self.section.optimization_steps = len(scc)

        if not self.section.input_energy_difference_tolerance:
            try:
                tolerance = self.section_run.section_sampling_method[-1].geometry_optimization_energy_change
                self.section.input_energy_difference_tolerance = tolerance
            except Exception:
                pass

        if not self.section.input_force_maximum_tolerance:
            try:
                tolerance = self.section_run.section_sampling_method[-1].geometry_optimization_threshold_force
                self.section.input_force_maximum_tolerance = tolerance
            except Exception:
                pass

        if not self.section.input_displacement_maximum_tolerance:
            try:
                tolerance = self.section_run.section_sampling_method[-1].geometry_optimization_geometry_change
                self.section.input_displacement_maximum_tolerance = tolerance
            except Exception:
                pass

        if not self.section.final_energy_difference:
            energies = []
            for scc in self.section_run.section_single_configuration_calculation:
                if scc.energy_total:
                    energies.append(scc.energy_total)

            delta_energy = resolve_difference(energies)
            if delta_energy is not None:
                self.section.final_energy_difference = delta_energy

        if not self.section.final_force_maximum:
            scc = self.section_run.section_single_configuration_calculation
            max_force = None
            if scc:
                if scc[-1].atom_forces is not None:
                    forces = self._to_numpy_array(scc[-1].atom_forces)
                    max_force = np.max(np.linalg.norm(forces, axis=1))
            if max_force is not None:
                self.section.final_force_maximum = max_force

        if not self.section.final_displacement_maximum:
            try:
                system = self.section_run.system
                displacements = [np.max(np.abs(
                    system[n].atom_positions - system[n - 1].atom_positions)) for n in range(1, len(system))]
                self.section.final_displacement_maximum = resolve_difference(displacements)
            except Exception:
                pass

        if not self.section.is_converged_geometry:
            # we can have several criteria for convergence: energy, force, displacement
            criteria = []
            try:
                criteria.append(self.section.final_energy_difference <= self.section.input_energy_difference_tolerance)
            except Exception:
                pass

            try:
                criteria.append(self.section.final_force_maximum <= self.section.input_force_maximum_tolerance)
            except Exception:
                pass

            try:
                criteria.append(self.section.final_displacement_maximum <= self.section.input_displacement_maximum_tolerance)
            except Exception:
                pass

            # converged when either criterion is met
            if criteria:
                self.section.is_converged_geometry = True in criteria


class PhononNormalizer(Normalizer):
    def __init__(self, entry_archive):
        super().__init__(entry_archive)

    def _get_n_imaginary_frequencies(self):
        scc = self.section_run.section_single_configuration_calculation
        if not scc:
            return
        sec_band = scc[0].section_k_band
        if not sec_band:
            return
        result = 0
        for band_segment in sec_band[0].section_k_band_segment:
            freq = band_segment.band_energies
            result += np.count_nonzero(np.array(freq) < 0)
        return result

    def normalize(self):
        self.section = self.entry_archive.section_workflow.section_phonon

        if not self.section:
            self.section = self.entry_archive.section_workflow.m_create(Phonon)

        if not self.section.n_imaginary_frequencies:
            # get number from bands (not complete as this is not the whole mesh)
            self.section.n_imaginary_frequencies = self._get_n_imaginary_frequencies()


class ElasticNormalizer(Normalizer):
    def __init__(self, entry_archive):
        super().__init__(entry_archive)

    def _resolve_mechanical_stability(self):
        spacegroup = self.entry_archive.section_run[-1].section_system[-1].x_elastic_space_group_number
        order = self.entry_archive.section_run[-1].section_method[-1].x_elastic_elastic_constant_order

        if spacegroup is None or order != 2:
            return

        scc = self.section_run.section_single_configuration_calculation[-1]
        c = scc.x_elastic_2nd_order_constants_matrix

        if c is None:
            return

        # see Phys. Rev B 90, 224104 (2014)
        res = False
        if spacegroup <= 2:  # Triclinic
            res = np.count_nonzero(c < 0)
        elif spacegroup <= 15:  # Monoclinic
            res = np.count_nonzero(c < 0)
        elif spacegroup <= 74:  # Orthorhombic
            res =\
                c[0][0] > 0 and c[0][0] * c[1][1] > c[0][1] ** 2 and\
                c[0][0] * c[1][1] * c[2][2] + 2 * c[0][1] * c[0][2] * c[1][2] -\
                c[0][0] * c[1][2] ** 2 - c[1][1] * c[0][2] ** 2 - c[2][2] * c[0][1] ** 2 > 0 and\
                c[3][3] > 0 and c[4][4] > 0 and c[5][5] > 0
        elif spacegroup <= 88:  # Tetragonal II
            res =\
                c[0][0] > abs(c[0][1]) and\
                2 * c[0][2] ** 2 < c[2][2] * (c[0][0] + c[0][1])
        elif spacegroup <= 142:  # Tetragonal I
            res =\
                c[0][0] > abs(c[0][1]) and\
                2 * c[0][2] ** 2 < c[2][2] * (c[0][0] + c[0][1]) and\
                c[3][3] > 0 and c[5][5] > 0
        elif spacegroup <= 148:  # rhombohedral II
            res =\
                c[0][0] > abs(c[0][1]) and c[3][3] > 0 and\
                c[0][2] ** 2 < (0.5 * c[2][2] * (c[0][0] + c[0][1])) and\
                c[0][3] ** 2 + c[0][4] ** 2 < 0.5 * c[3][3] * (c[0][0] - c[0][1])
        elif spacegroup <= 167:  # rhombohedral I
            res =\
                c[0][0] > abs(c[0][1]) and c[3][3] > 0 and\
                c[0][2] ** 2 < 0.5 * c[2][2] * (c[0][0] + c[0][1]) and\
                c[0][3] ** 2 < 0.5 * c[3][3] * (c[0][0] - c[0][1])
        elif spacegroup <= 194:  # hexagonal I
            res =\
                c[0][0] > abs(c[0][1]) and\
                2 * c[0][2] ** 2 < c[2][2] * (c[0][0] + c[0][1]) and\
                c[3][3] > 0 and c[5][5] > 0
        else:  # cubic
            res = c[0][0] - c[0][1] > 0 and c[0][0] + 2 * c[0][1] > 0 and c[3][3] > 0

        return res

    def _get_maximum_fit_error(self):
        scc = self.section_run.section_single_configuration_calculation[-1]

        max_error = 0.0
        for diagram in scc.x_elastic_section_strain_diagrams:
            if diagram.x_elastic_strain_diagram_type == 'cross-validation':
                error = np.amax(diagram.x_elastic_strain_diagram_values)
                max_error = error if error > max_error else max_error

        return max_error

    def normalize(self):
        self.section = self.entry_archive.section_workflow.section_elastic

        if not self.section:
            self.section = self.entry_archive.section_workflow.m_create(Elastic)

        if self.section.is_mechanically_stable is None:
            self.section.is_mechanically_stable = bool(self._resolve_mechanical_stability())

        if self.section.fitting_error_maximum is None:
            self.section.fitting_error_maximum = self._get_maximum_fit_error()


class MolecularDynamicsNormalizer(Normalizer):
    def __init__(self, entry_archive):
        super().__init__(entry_archive)

    def _is_with_thermodynamics(self):
        scc = self.section_run.section_single_configuration_calculation
        res = False
        if scc:
            res = scc[-1].temperature is not None
        return res

    def _is_with_trajectory(self):
        sec_system = self.section_run.section_system
        res = False
        if sec_system:
            res = sec_system[-1].atom_positions is not None
        return res

    def normalize(self):
        self.section = self.entry_archive.section_workflow.section_molecular_dynamics

        if not self.section:
            self.section = self.entry_archive.section_workflow.m_create(MolecularDynamics)

        if self.section.with_thermodynamics is None:
            self.section.with_thermodynamics = self._is_with_thermodynamics()

        if self.section.with_trajectory is None:
            self.section.with_trajectory = self._is_with_trajectory()


class WorkflowNormalizer(Normalizer):
    '''
    This normalizer produces information specific to a workflow.
    '''
    def __init__(self, entry_archive):
        super().__init__(entry_archive)
        self._elastic_programs = ['elastic']
        self._phonon_programs = ['phonopy']

    def _resolve_workflow_type_vasp(self):
        sec_method = self.section_run.section_method
        if not sec_method:
            return

        incar = self.section_run.section_method[0].x_vasp_incar_out
        nsw = incar.get('NSW')
        ibrion = -1 if nsw == 0 else incar.get('IBRION', 0)

        if ibrion == -1:
            return 'single_point'
        elif ibrion == 0:
            return 'molecular_dynamics'
        else:
            return 'geometry_optimization'

    def _resolve_workflow_type(self):
        # first get it from section_sampling_method
        workflow_type = None
        sec_sampling_method = self.section_run.section_sampling_method
        if sec_sampling_method:
            workflow_type = sec_sampling_method[-1].sampling_method
            # some parsers e.g. turbomole outputs geometry optimization
            if workflow_type in ['geometry optimization', 'relaxation']:
                workflow_type = 'geometry_optimization'

        # resolve it from parser
        if not workflow_type:
            program_name = self.section_run.program_name
            if program_name:
                program_name = program_name.lower()

            if program_name == 'vasp':
                workflow_type = self._resolve_workflow_type_vasp()

            elif program_name == 'elastic':
                workflow_type = 'elastic'

            elif program_name == 'lammps':
                workflow_type = 'molecular_dynamics'

            elif program_name == 'phonopy':
                workflow_type = 'phonon'

        # resolve if from scc
        if not workflow_type:
            if len(self.section_run.section_single_configuration_calculation) == 1:
                workflow_type = 'single_point'

        return workflow_type

    def normalize(self, logger=None) -> None:
        # Setup logger
        if logger is not None:
            self.logger = logger.bind(normalizer=self.__class__.__name__)

        # Do nothing if section_run is not present
        if self.section_run is None:
            return

        workflow_type = None
        if self.entry_archive.section_workflow:
            workflow_type = self.entry_archive.section_workflow.workflow_type

        if not workflow_type:
            workflow_type = self._resolve_workflow_type()

        if not workflow_type:
            return

        workflow = self.entry_archive.section_workflow
        if not workflow:
            workflow = self.entry_archive.m_create(Workflow)

        workflow.workflow_type = workflow_type

        if workflow.workflow_type == 'geometry_optimization':
            GeometryOptimizationNormalizer(self.entry_archive).normalize()

        elif workflow.workflow_type == 'phonon':
            PhononNormalizer(self.entry_archive).normalize()

        elif workflow.workflow_type == 'elastic':
            ElasticNormalizer(self.entry_archive).normalize()

        elif workflow.workflow_type == 'molecular_dynamics':
            MolecularDynamicsNormalizer(self.entry_archive).normalize()

        elif workflow.workflow_type == 'single_point':
            SinglePointNormalizer(self.entry_archive).normalize()

        scc = self.section_run.section_single_configuration_calculation
        if not self.entry_archive.section_workflow.calculation_result_ref:
            if scc:
                self.entry_archive.section_workflow.calculation_result_ref = scc[-1]

        if not self.entry_archive.section_workflow.calculations_ref:
            if scc:
                self.entry_archive.section_workflow.calculations_ref = scc

        # remove the section workflow again, if the parser/normalizer could not produce a result
        if workflow.calculation_result_ref is None:
            self.entry_archive.m_remove_sub_section(EntryArchive.section_workflow, -1)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='atk_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='atk.nomadmetainfo.json'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_atk_density_convergence_criterion = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Density convergence criteria to break the SCF cycle
        ''',
        a_legacy=LegacyDefinition(name='x_atk_density_convergence_criterion'))

    x_atk_mix_old = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of old densities in the density mixer
        ''',
        a_legacy=LegacyDefinition(name='x_atk_mix_old'))

    x_atk_mix_weight = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Mixing weight in density mixer
        ''',
        a_legacy=LegacyDefinition(name='x_atk_mix_weight'))

    x_atk_monkhorstpack_sampling = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        Monkhorstpack grid sampling
        ''',
        a_legacy=LegacyDefinition(name='x_atk_monkhorstpack_sampling'))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import datetime
import numpy as np

from nomad.datamodel import EntryArchive
from nomad.parsing import MatchingParser
from nomad.units import ureg as units
from nomad.datamodel.metainfo.public import section_run as Run
from nomad.datamodel.metainfo.public import section_system as System
from nomad.datamodel.metainfo.public import section_single_configuration_calculation as SCC

from nomad.parsing.file_parser import UnstructuredTextFileParser, Quantity

from . import metainfo  # pylint: disable=unused-import

'''
This is a hello world style example for an example parser/converter.
'''


def str_to_sites(string):
    sym, pos = string.split('(')
    pos = np.array(pos.split(')')[0].split(',')[:3], dtype=float)
    return sym, pos


calculation_parser = UnstructuredTextFileParser(quantities=[
    Quantity('sites', r'([A-Z]\([\d\.\, \-]+\))', str_operation=str_to_sites, repeats=True),
    Quantity(
        System.lattice_vectors,
        r'(?:latice|cell): \((\d)\, (\d), (\d)\)\,?\s*\((\d)\, (\d), (\d)\)\,?\s*\((\d)\, (\d), (\d)\)\,?\s*',
        repeats=False),
    Quantity('energy', r'energy: (\d\.\d+)'),
    Quantity('magic_source', r'done with magic source\s*\*{3}\s*\*{3}\s*[^\d]*(\d+)', repeats=False)])

mainfile_parser = UnstructuredTextFileParser(quantities=[
    Quantity('date', r'(\d\d\d\d\/\d\d\/\d\d)', repeats=False),
    Quantity('program_version', r'super\_code\s*v(\d+)\s*', repeats=False),
    Quantity(
        'calculation', r'\s*system \d+([\s\S]+?energy: [\d\.]+)([\s\S]+\*\*\*)*',
        sub_parser=calculation_parser,
        repeats=True)
])


class ExampleParser(MatchingParser):
    def __init__(self):
        super().__init__(
            name='parsers/example', code_name='EXAMPLE', code_homepage='https://www.example.eu/',
            mainfile_mime_re=r'(application/.*)|(text/.*)',
            mainfile_contents_re=(r'^\s*#\s*This is example output'),
            supported_compressions=['gz', 'bz2', 'xz']
        )

    def parse(self, mainfile: str, archive: EntryArchive, logger):
        # Log a hello world, just to get us started. TODO remove from an actual parser.
        logger.info('Hello World')

        # Use the previously defined parsers on the given mainfile
        mainfile_parser.mainfile = mainfile
        mainfile_parser.parse()

        # Output all parsed data into the given archive.
        run = archive.m_create(Run)
        run.program_name = 'super_code'
        run.program_version = str(mainfile_parser.get('program_version'))
        date = datetime.datetime.strptime(
            mainfile_parser.get('date'),
            '%Y/%m/%d') - datetime.datetime(1970, 1, 1)
        run.program_compilation_datetime = date.total_seconds()

        for calculation in mainfile_parser.get('calculation'):
            system = run.m_create(System)

            system.lattice_vectors = calculation.get('lattice_vectors')
            sites = calculation.get('sites')
            system.atom_labels = [site[0] for site in sites]
            system.atom_positions = [site[1] for site in sites]

            scc = run.m_create(SCC)
            scc.single_configuration_calculation_to_system_ref = system
            scc.energy_total = calculation.get('energy') * units.eV
            scc.single_configuration_calculation_to_system_ref = system
            magic_source = calculation.get('magic_source')
            if magic_source is not None:
                scc.x_example_magic_value = magic_source



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class FploParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*\-*\+*\d*'

        super().__init__(
            specifications=dict(
                name='parsers/fplo', code_name='fplo', domain='dft',
                mainfile_contents_re=r'\s*\|\s*FULL-POTENTIAL LOCAL-ORBITAL MINIMUM BASIS BANDSTRUCTURE CODE\s*\|\s*',
                mainfile_mime_re=r'text/.*'),
            units_mapping=dict(length=ureg.bohr, energy=ureg.eV),
            program_version=r'main version\:\s*(\S+)[\|\s]+sub  version\:\s*(\S+)[\|\s]+release\s*\:\s*(\S+)',
            lattice_vectors=r'lattice vectors\s*(a1\s*\:\s*[\s\S]+?)rec',
            atom_labels_atom_positions=rf'No\. *Element WPS CPA\-Block *X *Y *Z([\s\S]+?)\n *\n',
            energy_reference_fermi=(rf'Fermi energy\:\s*({re_f}).+electrons', lambda x: [x]),
            energy_total=rf'total energy.+\s*EE\:\s*({re_f})')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='crystal_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='crystal.nomadmetainfo.json'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_crystal_family = Quantity(
        type=str,
        shape=[],
        description='''
        Crystal family.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_family'))

    x_crystal_class = Quantity(
        type=str,
        shape=[],
        description='''
        Crystal class.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_class'))

    x_crystal_spacegroup = Quantity(
        type=str,
        shape=[],
        description='''
        Crystal spacegroup string resembling Hermann–Mauguin notation.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_spacegroup'))

    x_crystal_dimensionality = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        System dimensionality.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_dimensionality'))

    x_crystal_n_symmops = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of symmetry operators.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_number_of_symmops'))


class section_scf_iteration(public.section_scf_iteration):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_scf_iteration'))

    x_crystal_scf_ee = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        +++ ENERGIES IN A.U. +++. ::: TOTAL E-E
        4.6595142576204E+01
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_scf_total_ee'))

    x_crystal_scf_en_ne = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        +++ ENERGIES IN A.U. +++. ::: TOTAL E-N + N-E
        -5.2283101954878E+02
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_scf_total_en_ne'))

    x_crystal_scf_nn = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        +++ ENERGIES IN A.U. +++. ::: TOTAL N-N
        -7.3084276676762E+01
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_scf_total_nn'))

    x_crystal_scf_virial_coefficient = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        +++ ENERGIES IN A.U. +++. ::: VIRIAL COEFFICIENT
        9.9998501747632E-01
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_scf_virial_coefficient'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_crystal_run_title = Quantity(
        type=str,
        shape=[],
        description='''
        Title of the runcry(14) task.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_run_title'))

    x_crystal_datetime = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary type for storing date and time, in locale-dependent format.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_process_datetime'))

    x_crystal_executable_path = Quantity(
        type=str,
        shape=[],
        description='''
        Crystal executable filepath.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_process_exe'))

    x_crystal_hostname = Quantity(
        type=str,
        shape=[],
        description='''
        Hostname where Crystal was run.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_process_hn'))

    x_crystal_input_path = Quantity(
        type=str,
        shape=[],
        description='''
        Input file name.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_process_input'))

    x_crystal_os = Quantity(
        type=str,
        shape=[],
        description='''
        String describing the operating system where Crystal was run.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_process_os'))

    x_crystal_output = Quantity(
        type=str,
        shape=[],
        description='''
        Output file name.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_process_output'))

    x_crystal_tmpdir = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary directory.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_process_tmpdir'))

    x_crystal_user = Quantity(
        type=str,
        shape=[],
        description='''
        Username: who ran Crystal.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_process_user'))

    x_crystal_distribution = Quantity(
        type=str,
        shape=[],
        description='''
        Distribution describer.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_header_distribution'))

    x_crystal_version_minor = Quantity(
        type=str,
        shape=[],
        description='''
        Minor version number of Crystal.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_header_minor'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_crystal_convergence_deltap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Convergence seettings, on power of 10 (e.g. CONVERGENCE ON DELTAP        10**-16)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_convergence_on_deltap'))

    x_crystal_weight_f = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        WEIGHT OF F(I) IN F(I+1)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_properties_weight_f'))

    x_crystal_coulomb_bipolar_buffer = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        COULOMB BIPOLAR BUFFER SET TO x Mb
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_coulomb_bipolar_buffer'))

    x_crystal_eigenvectors_disk_space_ftn = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        DISK SPACE FOR EIGENVECTORS (FTN 10)      351575 REALS
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_eigenvectors_disk_space_ftn'))

    x_crystal_eigenvectors_disk_space_reals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        DISK SPACE FOR EIGENVECTORS (FTN 10)      351575 REALS
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_eigenvectors_disk_space_reals'))

    x_crystal_exchange_bipolar_buffer = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        EXCHANGE BIPOLAR BUFFER SET TO x Mb
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_exchange_bipolar_buffer'))

    x_crystal_fock_ks_matrix_mixing = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        FOCK/KS MATRIX MIXING SET TO x %
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_fock_ks_matrix_mixing'))

    x_crystal_input_tcpu = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TTTTTTTTTTTTTTTTTTTTTTTTTTTTTT INPUT       TELAPSE        0.01 TCPU        0.01
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_input_tcpu'))

    x_crystal_input_telapse = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TTTTTTTTTTTTTTTTTTTTTTTTTTTTTT INPUT       TELAPSE        0.01 TCPU        0.01
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_input_telapse'))

    x_crystal_irr_f = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        MATRIX SIZE: P(G)   31533, F(G)    5204, P(G) IRR    1802, F(G) IRR     964
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_irr_f'))

    x_crystal_irr_p = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        MATRIX SIZE: P(G)   31533, F(G)    5204, P(G) IRR    1802, F(G) IRR     964
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_irr_p'))

    x_crystal_is1 = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        CAPPA:IS1 16;IS2 16;IS3 16; K PTS MONK NET 145; SYMMOPS:K SPACE  48;G SPACE  48.
        (mentioned after the basis set)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_is1'))

    x_crystal_is2 = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        CAPPA:IS1 16;IS2 16;IS3 16; K PTS MONK NET 145; SYMMOPS:K SPACE  48;G SPACE  48.
        (mentioned after the basis set)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_is2'))

    x_crystal_is3 = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        CAPPA:IS1 16;IS2 16;IS3 16; K PTS MONK NET 145; SYMMOPS:K SPACE  48;G SPACE  48.
        (mentioned after the basis set)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_is3'))

    x_crystal_n_k_points_gilat = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        NUMBER OF K POINTS(GILAT NET)    145
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_k_points_gilat'))

    x_crystal_n_k_points_ibz = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        NUMBER OF K POINTS IN THE IBZ    145
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_k_points_ibz'))

    x_crystal_k_pts_monk_net = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        CAPPA:IS1 16;IS2 16;IS3 16; K PTS MONK NET 145; SYMMOPS:K SPACE  48;G SPACE  48.
        (mentioned after the basis set)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_k_pts_monk_net'))

    x_crystal_matrix_size_f = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        MATRIX SIZE: P(G)   31533, F(G)    5204, P(G) IRR    1802, F(G) IRR     964
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_matrix_size_f'))

    x_crystal_matrix_size_p = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        MATRIX SIZE: P(G)   31533, F(G)    5204, P(G) IRR    1802, F(G) IRR     964
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_matrix_size_p'))

    x_crystal_max_g_vector_index = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        MAX G-VECTOR INDEX FOR 1- AND 2-ELECTRON INTEGRALS 247
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_max_g_vector_index'))

    x_crystal_max_scf_cycles = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        MAX NUMBER OF SCF CYCLES
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_max_scf_cycles'))

    x_crystal_n_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        N. OF ATOMS PER CELL
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_number_of_atoms'))

    x_crystal_n_core_electrons = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        CORE ELECTRONS PER CELL
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_number_of_core_electrons'))

    x_crystal_n_electrons = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        N. OF ELECTRONS PER CELL
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_number_of_electrons'))

    x_crystal_n_orbitals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        NUMBER OF AO (Atomic orbitals)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_number_of_orbitals'))

    x_crystal_n_shells = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        NUMBER OF SHELLS
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_number_of_shells'))

    x_crystal_n_symmops = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        N. OF SYMMETRY OPERATORS
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_number_of_symmops'))

    x_crystal_pole_order = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        POLE ORDER IN MONO ZONE
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_pole_order'))

    x_crystal_shrink_gilat = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        SHRINKING FACTOR(GILAT NET)   16
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_shrink_gilat'))

    x_crystal_shrink_value1 = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_shrink_value1'))

    x_crystal_shrink_value2 = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_shrink_value2'))

    x_crystal_shrink_value3 = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_shrink_value3'))

    x_crystal_shrink = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        SHRINK. FACT.(MONKH.)   16 16 16
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_shrink'))

    x_crystal_symmetry_adaption = Quantity(
        type=bool,
        shape=[],
        description='''
        SYMMETRY ADAPTION OF THE BLOCH FUNCTIONS ENABLED
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_symmetry_adaption'))

    x_crystal_symmops_g = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        CAPPA:IS1 16;IS2 16;IS3 16; K PTS MONK NET 145; SYMMOPS:K SPACE  48;G SPACE  48.
        (mentioned after the basis set)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_symmops_g'))

    x_crystal_symmops_k = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        CAPPA:IS1 16;IS2 16;IS3 16; K PTS MONK NET 145; SYMMOPS:K SPACE  48;G SPACE  48.
        (mentioned after the basis set)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_symmops_k'))

    x_crystal_tol_coulomb_overlap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        COULOMB OVERLAP TOL         (T1) 10**   -6. (Tolerance T1, power of 10)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_tol_coulomb_overlap'))

    x_crystal_tol_coulomb_penetration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        COULOMB PENETRATION TOL     (T2) 10**   -6. (Tolerance T2, power of 10)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_tol_coulomb_penetration'))

    x_crystal_tol_exchange_overlap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        EXCHANGE OVERLAP TOL        (T3) 10**   -6. (Tolerance T3, power of 10)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_tol_exchange_overlap'))

    x_crystal_tol_pseudo_overlap_f = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        EXCHANGE PSEUDO OVP (F(G))  (T4) 10**   -6. (Tolerance T4, power of 10)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_tol_pseudo_overlap_f'))

    x_crystal_tol_pseudo_overlap_p = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        EXCHANGE PSEUDO OVP (P(G))  (T5) 10**  -12. (Tolerance T5, power of 10)
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_tol_pseudo_overlap_p'))

    x_crystal_type_of_calculation = Quantity(
        type=str,
        shape=[],
        description='''
        The type of the calculation that was performed.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_type_of_calculation'))

    x_crystal_weight_previous = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        WEIGHT OF F(I) IN F(I+1)      30%
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_weight_previous'))

    x_crystal_toldee = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TOLDEE info
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_info_toldee'))


class section_basis_set_atom_centered(public.section_basis_set_atom_centered):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_basis_set_atom_centered'))

    x_crystal_section_shell = SubSection(
        sub_section=SectionProxy('x_crystal_section_shell'),
        repeats=True
    )


class x_crystal_section_shell(MSection):
    '''
    Shell contains a number of orbitals.
    '''
    m_def = Section(validate=False)

    x_crystal_shell_range = Quantity(
        type=str,
        shape=[],
        description='''
        The range of orbitals
        ''')

    x_crystal_shell_type = Quantity(
        type=str,
        shape=[],
        description='''
        Shell type: S / P / SP / D / F / G.
        ''',
        a_legacy=LegacyDefinition(name='x_crystal_basis_set_atom_shell_type'))

    x_crystal_shell_coefficients = Quantity(
        type=np.dtype(np.float64),
        shape=['n_orbitals', 4],
        description='''
        Contraction coefficients in this order: exponent, S, P, D/F/G.
        '''
    )


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
from nomad.metainfo import Environment
from nomad.metainfo.legacy import LegacyMetainfoEnvironment
import namdparser.metainfo.namd
import nomad.datamodel.metainfo.common
import nomad.datamodel.metainfo.public
import nomad.datamodel.metainfo.general

m_env = LegacyMetainfoEnvironment()
m_env.m_add_sub_section(Environment.packages, sys.modules['namdparser.metainfo.namd'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.metainfo.common'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.metainfo.public'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.metainfo.general'].m_package)  # type: ignore



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import numpy as np

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class MopacParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        def get_forces(val):
            try:
                forces = re.findall(rf'({re_f}) +KCAL\/ANGSTROM', val)
                return np.reshape(np.array(
                    forces, dtype=np.dtype(np.float64)), (len(forces) // 3, 3)) * (ureg.kcal / ureg.angstrom / 6.02214076e+23)
            except Exception:
                return []

        super().__init__(
            specifications=dict(
                name='parsers/mopac', code_name='MOPAC', domain='dft',
                mainfile_contents_re=r'\s*\*\*\s*MOPAC\s*([0-9a-zA-Z]*)\s*\*\*\s*',
                mainfile_mime_re=r'text/.*'),
            units_mapping=dict(length=ureg.angstrom, energy=ureg.eV),
            # include code name to distinguish gamess and firefly
            program_version=r'Version ([\w\.]+)',
            atom_labels_atom_positions=r'CARTESIAN COORDINATES\s*(1[\s\S]+?)\n *\n',
            energy_total=rf'TOTAL ENERGY *\= *({re_f})',
            atom_forces=(r'TYPE +VALUE +GRADIENT\s*([\s\S]+?)\n *\n', get_forces))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

''' Temporary code needed for ops, version transitions, etc. '''

from nomad.datamodel import EntryMetadata
from nomad.normalizing.optimade import transform_to_v1 as optimade


external_dbs = {
    'AFLOW': '81b96683-7170-49d7-8c4e-e9f34906b3ea',
    'Materials Project': 'ab0305be-b3bb-41e6-a000-b004b49e475a',
    'OQMD': '9ac9db26-b268-4c22-b3f7-fc82ce35967e'
}
external_dbs_from_user_id = {value: key for key, value in external_dbs.items()}


def transform_to_v0_10(entry: EntryMetadata) -> EntryMetadata:
    entry = optimade(entry)

    # external db
    uploader_id = entry.uploader.user_id
    if uploader_id in external_dbs_from_user_id and entry.external_db is None:
        entry.external_db = external_dbs_from_user_id[uploader_id]

    return entry


def set_external_db_in_mongo(upload_id: str = None):
    from nomad import infrastructure
    from nomad.processing import Calc
    infrastructure.setup_mongo()

    calcs = Calc._get_collection()
    print(calcs)

    for external_db, user_id in external_dbs.items():
        query = {
            'metadata.uploader': user_id,
            'metadata.external_db': {
                '$exists': False
            }}
        if upload_id is not None:
            query['upload_id'] = upload_id

        update = {
            '$set': {
                'metadata.external_db': external_db
            }
        }

        result = calcs.update_many(query, update)
        print(
            f'updated for {external_db}, update acknowledged ({result.acknowledged}), '
            f'matches = {result.matched_count}')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import numpy as np

from nomad.parsing.file_parser import BasicParser
from nomad.units import ureg


class CharmmParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        def get_positions(val):
            res = dict(atom_labels=[], atom_positions=[])
            try:
                labels, positions = zip(*re.findall(rf'([A-Z])\w* +({re_f} +{re_f} +{re_f}) +\w', val))
                res['atom_labels'] = labels
                res['atom_positions'] = np.array([v.split() for v in positions], dtype=np.dtype(np.float64)) * ureg.angstrom
            except Exception:
                raise
                pass
            return res

        super().__init__(
            specifications=dict(
                name='parsers/charmm', code_name='Charmm', domain='dft',
                mainfile_contents_re=r'\s*Chemistry\s*at\s*HARvard\s*Macromolecular\s*Mechanics\s*',
                mainfile_mime_re=r'text/.*'),
            units_mapping=dict(length=ureg.angstrom, energy=ureg.kcal / 6.02214076e+23),
            auxilliary_files=r'\/*([a-z][\w\-]+.crd)',
            program_version=r'\(CHARMM\) \- (.+Version[\w \,]+?\d\d\d\d)',
            atom_labels_atom_positions=(
                rf'(\*.*\s*\*.*\s*\*.*\s*\d+\n +\d+ +\d+ +\w+ +[A-Z]\w* +{re_f} +{re_f} +{re_f}[\s\S]+)(?:\>|\Z|\n *\n)',
                get_positions),
            energy_total=rf'\n *[A-Z]+\> +\d* +({re_f})')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='fplo_temporaries_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='fplo.temporaries.nomadmetainfo.json'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_fplo_t_program_version_main = Quantity(
        type=str,
        shape=[],
        description='''
        temporary: FPLO main version
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_program_version_main'))

    x_fplo_t_program_version_release = Quantity(
        type=str,
        shape=[],
        description='''
        temporary: FPLO release number
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_program_version_release'))

    x_fplo_t_run_hosts = Quantity(
        type=str,
        shape=[],
        description='''
        temporary: FPLO run hosts
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_run_hosts'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_fplo_t_vec_a_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for direct lattice vectors, x-component
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_vec_a_x'))

    x_fplo_t_vec_a_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for direct lattice vectors, y-component
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_vec_a_y'))

    x_fplo_t_vec_a_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for direct lattice vectors, z-component
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_vec_a_z'))

    x_fplo_t_vec_b_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for reciprocal lattice vectors, x-component
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_vec_b_x'))

    x_fplo_t_vec_b_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for reciprocal lattice vectors, y-component
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_vec_b_y'))

    x_fplo_t_vec_b_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for reciprocal lattice vectors, z-component
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_vec_b_z'))

    x_fplo_t_atom_positions_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for atom positions, x-component
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_atom_positions_x'))

    x_fplo_t_atom_positions_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for atom positions, y-component
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_atom_positions_y'))

    x_fplo_t_atom_positions_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for atom positions, z-component
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_atom_positions_z'))

    x_fplo_t_atom_idx = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO atom index
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_atom_idx'))

    x_fplo_t_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO atom labels
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_atom_labels'))

    x_fplo_t_atom_wyckoff_idx = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO Wyckoff position index of each atom
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_atom_wyckoff_idx'))

    x_fplo_t_atom_cpa_block = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO CPA block of each atom
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_atom_cpa_block'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_fplo_t_relativity_method = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO relativistic method
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_relativity_method'))

    x_fplo_t_dft_plus_u_species_subshell_species = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U species
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_species_subshell_species'))

    x_fplo_t_dft_plus_u_species_subshell_element = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U element
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_species_subshell_element'))

    x_fplo_t_dft_plus_u_species_subshell_subshell = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U (n,l)subshell
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_species_subshell_subshell'))

    x_fplo_t_dft_plus_u_species_subshell_F0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U F0
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_species_subshell_F0'))

    x_fplo_t_dft_plus_u_species_subshell_F2 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U F2
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_species_subshell_F2'))

    x_fplo_t_dft_plus_u_species_subshell_F4 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U F4
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_species_subshell_F4'))

    x_fplo_t_dft_plus_u_species_subshell_F6 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U F6
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_species_subshell_F6'))

    x_fplo_t_dft_plus_u_species_subshell_U = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U U
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_species_subshell_U'))

    x_fplo_t_dft_plus_u_species_subshell_J = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Temporary storage for FPLO per species/(n,l)subshell DFT+U J
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_species_subshell_J'))

    x_fplo_t_dft_plus_u_site_index = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U index
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_site_index'))

    x_fplo_t_dft_plus_u_site_element = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U element
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_site_element'))

    x_fplo_t_dft_plus_u_site_species = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U species
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_site_species'))

    x_fplo_t_dft_plus_u_site_subshell = Quantity(
        type=str,
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U (n,l)subshell
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_site_subshell'))

    x_fplo_t_dft_plus_u_site_ubi1 = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U ubi1
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_site_ubi1'))

    x_fplo_t_dft_plus_u_site_ubi2 = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Temporary storage for FPLO per site DFT+U ubi2
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_t_dft_plus_u_site_ubi2'))


class section_scf_iteration(public.section_scf_iteration):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_scf_iteration'))

    x_fplo_t_energy_reference_fermi_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Temporary storage for FPLO Fermi energy in iteration
        ''',
        categories=[public.energy_type_reference, public.energy_value],
        a_legacy=LegacyDefinition(name='x_fplo_t_energy_reference_fermi_iteration'))


m_package.__init_metainfo__()
